/* 
 * File:   WinrtFileMgr.cpp
 * Author: Tiago
 * 
 * Created on May 30, 2010, 4:44 PM
 */

#include "WinrtFileMgr.h"
#include "1-texts/wchar_utils.h"
#include "1-texts/TextMgr.h"
#include "2-application/MemoryMgr.h"
#include "1-files/WinrtPersistentMgr.h"
#include "1-files/WinrtFile.h"

//#include <Windows.h>
#include "2-application/Application.h"
#include "2-store/WinrtDLCMgr.h"

#if (WINAPI_FAMILY == WINAPI_FAMILY_APP)
#include "1-files/WinrtFileDownloader.h"
#endif

#if (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#include "1-files/WP8FileDownloaderWrapper.h"
#include "1-files/WP8StorageWrapper.h"
#endif

#include <ppltasks.h>

#include <wrl.h>


#include <mmreg.h>
#include <mfidl.h>
#include <mfapi.h>

#include <mfmediaengine.h>

using namespace Windows::Storage;
using namespace Windows::ApplicationModel;
using namespace std;

using namespace Pasta;

WinrtFileMgr *WinrtFileMgr::currentFileMgr = NULL;

// record the resource path
char s_pszResourcePath[MAX_PATH] = {0};


std::wstring WinrtFileMgr::CCUtf8ToUnicode(const char * pszUtf8Str, unsigned len/* = -1*/)
{
    std::wstring ret;
    do
    {
        if (! pszUtf8Str) break;
		// get UTF8 string length
		if (-1 == len)
		{
			len = strlen(pszUtf8Str);
		}
        if (len <= 0) break;

		// get UTF16 string length
		int wLen = MultiByteToWideChar(CP_UTF8, 0, pszUtf8Str, len, 0, 0);
		if (0 == wLen || 0xFFFD == wLen) break;
		
		// convert string  
        wchar_t * pwszStr = new wchar_t[wLen + 1];
        if (! pwszStr) break;
        pwszStr[wLen] = 0;
        MultiByteToWideChar(CP_UTF8, 0, pszUtf8Str, len, pwszStr, wLen + 1);
        ret = pwszStr;
        CC_SAFE_DELETE_ARRAY(pwszStr);
    } while (0);
    return ret;
}

std::string WinrtFileMgr::CCUnicodeToUtf8(const wchar_t* pwszStr)
{
	std::string ret;
	do
	{
		if(! pwszStr) break;
		size_t len = wcslen(pwszStr);
		if (len <= 0) break;
		
		size_t convertedChars = 0;
		char * pszUtf8Str = new char[len*3 + 1];
		WideCharToMultiByte(CP_UTF8, 0, pwszStr, len+1, pszUtf8Str, len*3 + 1, 0, 0);
		ret = pszUtf8Str;
		CC_SAFE_DELETE_ARRAY(pszUtf8Str);
	}while(0);

	return ret;
}
void WinrtFileMgr::_CheckPath()
{
	if (! s_pszResourcePath[0])
	{
		//WCHAR  wszPath[MAX_PATH];
		//int nNum = WideCharToMultiByte(CP_ACP, 0, wszPath, 
		//	GetCurrentDirectoryW(sizeof(wszPath), wszPath), 
		//	s_pszResourcePath, MAX_PATH, NULL, NULL);
		//      s_pszResourcePath[nNum] = '\\';

		Windows::ApplicationModel::Package^ package = Windows::ApplicationModel::Package::Current;
		Windows::Storage::StorageFolder^ installedLocation = package->InstalledLocation;
		Platform::String^ resPath = installedLocation->Path + "\\";
		std::string pathStr = CCUnicodeToUtf8(resPath->Data());
		strcpy_s(s_pszResourcePath, pathStr.c_str());
	}
}




std::string WinrtFileMgr::fullPathFromRelativePath(const char *pszRelativePath)
{
	_CheckPath();

    std::string pRet;
    if ((strlen(pszRelativePath) > 1 && pszRelativePath[1] == ':'))
    {
        // path start with "x:", is absolute path
        pRet = pszRelativePath;
    }
    else if (strlen(pszRelativePath) > 0 
        && ('/' == pszRelativePath[0] || '\\' == pszRelativePath[0]))
    {
        // path start with '/' or '\', is absolute path without driver name
		char szDriver[3] = {s_pszResourcePath[0], s_pszResourcePath[1], 0};
        pRet = szDriver;
        pRet += pszRelativePath;
    }
    else
    {
        pRet = s_pszResourcePath;
        pRet += pszRelativePath;
    }

	return pRet;
}

const char *WinrtFileMgr::fullPathFromRelativeFile(const char *pszFilename, const char *pszRelativeFile)
{
	_CheckPath();
	// std::string relativeFile = fullPathFromRelativePath(pszRelativeFile);
	std::string relativeFile = pszRelativeFile;
	string pRet;// = new string;
	pRet = relativeFile.substr(0, relativeFile.find_last_of("/\\") + 1);
	pRet += pszFilename;
	return pRet.c_str();
}


WinrtFileMgr::WinrtFileMgr()
{
	const int pathLen = 1024;
	wchar_t* completePathToFile = new wchar_t[pathLen];
// jp
//	if (SHGetSpecialFolderPathW(NULL, completePathToFile, CSIDL_MYDOCUMENTS, FALSE) != TRUE)
//	{
//		SAFE_DELETE_ARRAY_NULLIFY(completePathToFile);
//		return;
//	}

	/// Concatenate to %myDocuments%
	/*wcscat_s(completePathToFile, pathLen, L"\\Pastagames\\");
	WinrtFileMgr::CreateDir(completePathToFile);

	wcscat_s(completePathToFile, pathLen, L"Burn it All\\");
	WinrtFileMgr::CreateDir(completePathToFile);*/

}

WinrtFileMgr::~WinrtFileMgr() {


}

void WinrtFileMgr::init()
{

}

File* WinrtFileMgr::createFile(const std::string& fileName)
{
	return new WinrtFile(fileName);
}

char*		WinrtFileMgr::convertToWriteAuthorizedPath(const char* path)
{
	wchar_t* convertedW = this->convertToWriteAuthorizedPathW(path);
	char* converted = WideCharToAscii(convertedW);
	delete[] convertedW;
	return converted;
}


void *WinrtFileMgr::load(const wchar_t *path, u32 *fileSize, int upToNBytes, FILE* fileToRead)
{
	void * memory = NULL;

	if( fileToRead == NULL )
	{
		PastaLogDebug("GLWFileMgr: Can not open file\n");
		return NULL;
	}

	if (upToNBytes == -1)
	{
		fseek( fileToRead, 0, SEEK_END );
		*fileSize = ftell( fileToRead );
		fseek( fileToRead, 0, SEEK_SET );
	}
	else
	{
		*fileSize = upToNBytes;
	}

	memory = (void*) new char[*fileSize + 1];
	((char*)memory)[*fileSize] = 0;

	fread( memory, 1, *fileSize, fileToRead );

	if(ferror(fileToRead))
	{
		delete[] (char*)memory;
		PastaLogDebug("GLWFileMgr: Error reading file\n");
		return NULL;
	}

	fclose( fileToRead );

	return memory;
}

void* WinrtFileMgr::load(const std::string& path, u32 *fileSize )
{
	if (MemoryMgr::getSingleton())
	{
		PastaLogDebug("loading file %s\testimated used RAM: %dMo\n", path.c_str(), (MemoryMgr::getSingleton()->getCurrentMemoryUsage() / 1024 / 1024));
	}
	else
	{
		PastaLogDebug("loading file %s\n", path.c_str());
	}

	void * memory = NULL;

	FILE* fileToRead;
    string pszPath = fullPathFromRelativePath(path.c_str());
	fopen_s(&fileToRead, pszPath.c_str(), "rb");
	if( fileToRead == NULL )
	{
		// If not found, try to load it from the DLC folder.
		string dlcPath = WinrtDLCMgr::getDLCFolderPath();
		dlcPath.append(filenameFromPath(pszPath));
		fopen_s(&fileToRead, dlcPath.c_str(), "rb");
		if( fileToRead == NULL )
		{
			return NULL;
		}
		else
		{
			PastaLogDebug("DLC load file: %s\n", dlcPath.c_str());
		}
	}

	fseek( fileToRead, 0, SEEK_END );
	u32 size = ftell( fileToRead );
	if (fileSize) 
	{
		*fileSize = size;
	}
	fseek( fileToRead, 0, SEEK_SET );

	memory = (void*) new char[size + 1];

	fread( memory, 1, size, fileToRead );
	((char*)memory)[size] = 0;

	if(ferror(fileToRead))
	{
		assert(false);
		delete[] (char*)memory;
		PastaLogDebug("WinrtFileMgr: Error reading file %s", path.c_str());

		return NULL;
	}

	fclose( fileToRead );

	return memory;
}

void *WinrtFileMgr::load(const char *pszFileName, u32 *fileSize )
{
	_CheckPath();
    string pszPath = fullPathFromRelativePath(pszFileName);

  FILE_STANDARD_INFO fileStandardInfo = { 0 };
	HANDLE hFile;
	DWORD bytesRead = 0;
	uint32 dwFileSize = 0;
	BYTE* pBuffer = 0;

	std::wstring path = CCUtf8ToUnicode(pszPath.c_str(),-1);


	do 
	{
		CREATEFILE2_EXTENDED_PARAMETERS extendedParams = {0};
		extendedParams.dwSize = sizeof(CREATEFILE2_EXTENDED_PARAMETERS);
		extendedParams.dwFileAttributes = FILE_ATTRIBUTE_NORMAL;
		extendedParams.dwFileFlags = FILE_FLAG_SEQUENTIAL_SCAN;
		extendedParams.dwSecurityQosFlags = SECURITY_ANONYMOUS;
		extendedParams.lpSecurityAttributes = nullptr;
		extendedParams.hTemplateFile = nullptr;

		// read the file from hardware
		hFile = ::CreateFile2(path.c_str(), GENERIC_READ, FILE_SHARE_READ, OPEN_EXISTING, &extendedParams);
		if (INVALID_HANDLE_VALUE == hFile)
		{
			break;
		}


		BOOL result = ::GetFileInformationByHandleEx(
			hFile,
			FileStandardInfo,
			&fileStandardInfo,
			sizeof(fileStandardInfo)
			);

		//Read error
		if ((result == 0) || (fileStandardInfo.EndOfFile.HighPart != 0))
		{
			break;
		}

		dwFileSize = fileStandardInfo.EndOfFile.LowPart;
		//for read text
		pBuffer = new BYTE[dwFileSize+1];
		pBuffer[dwFileSize] = 0;
		if (!ReadFile(hFile, pBuffer, dwFileSize, &bytesRead, nullptr))
		{
			break;
		}
		*fileSize = bytesRead;

	} while (0);

	if (! pBuffer )
	{
		std::string title = "Notification";
		std::string msg = "WinrtFileMgr::load - Get data from file(";
		msg.append(pszPath).append(") failed!\n");

		PastaLogDebug(msg.c_str());
	}

    if (hFile != INVALID_HANDLE_VALUE)
        CloseHandle(hFile);

	return pBuffer;
}

void WinrtFileMgr::release(void *memory )
{
    	if (memory != NULL)
		delete[] (char*)memory;
}
void WinrtFileMgr::close()
{

}

bool	WinrtFileMgr::fileExists(const char* resPath)
{
    string pszPath = fullPathFromRelativePath(resPath);
    bool ret = false;
    FILE * pf = 0;
    const char * pathChar = pszPath.c_str();
    if (pathChar && strlen(pathChar) && (pf = fopen(pathChar, "rb")))
    {
        return true;
        fclose(pf);
    }
    WinrtPersistentMgr* persistentMgr = (WinrtPersistentMgr*)PersistentMgr::getSingleton();
    return  persistentMgr->fileExists(resPath);
}


int WinrtFileMgr::asynchronouslyDownloadContentOfURL(const std::string& URL, const std::string& destinationFilename)
{
#if (WINAPI_FAMILY == WINAPI_FAMILY_APP)
	WinrtFileDownloader::getDownloader()->downloadFile(stops(URL), stops(destinationFilename));
#endif
#if (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
	WP8FileDownloaderWrapper::getDelegate()->AddBackgroundTransfer(stops(URL), stops(destinationFilename));
#endif
	return 0;
}

int WinrtFileMgr::asynchronouslyDownloadContentOfURL(const std::string& URL) {
#if (WINAPI_FAMILY == WINAPI_FAMILY_APP)
    WinrtFileDownloader::getDownloader()->downloadFile(stops(URL));
#endif
#if (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
	WP8FileDownloaderWrapper::getDelegate()->AddBackgroundTransfer(stops(URL), "");
#endif
    return 0;
}

int WinrtFileMgr::getAsynchronousDownloadCompletion(const std::string& name) {
#if (WINAPI_FAMILY == WINAPI_FAMILY_APP)
	return WinrtFileDownloader::getDownloader()->fileProgress(stops(name));
#endif
#if (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
	return WP8FileDownloaderWrapper::getDelegate()->getProgression(stops(name));
#endif
	return 0;
}

Platform::String^ WinrtFileMgr::filenameFromPath(Platform::String^ path)
{
	std::wstring ws_url = path->Data();
	std::string ws_filename;
	auto it = ws_url.rbegin();
	auto it_end = ws_url.rend();
	while(it != it_end) {
		wchar_t c = *it;
		if(c=='\\' || c=='/') break;
		++it;
	}
	ws_filename.insert(ws_filename.begin(), ws_url.rbegin(), it);
	std::reverse(ws_filename.begin(), ws_filename.end());
	return atops(ws_filename.c_str());
}

Platform::String^ WinrtFileMgr::extensionFromPath(Platform::String^ path)
{
	std::wstring ws_url = path->Data();
	std::string ws_filename;
	auto it = ws_url.rbegin();
	auto it_end = ws_url.rend();
	while(it != it_end) {
		wchar_t c = *it;
		if(c=='.') break;
		++it;
	}
	ws_filename.insert(ws_filename.begin(), ws_url.rbegin(), it);
	std::reverse(ws_filename.begin(), ws_filename.end());
	return atops(ws_filename.c_str());
}

std::string WinrtFileMgr::filenameFromPath(std::string path)
{
	std::string ret;
	auto it = path.rbegin();
	auto it_end = path.rend();
	while(it != it_end) {
		wchar_t c = *it;
		if(c=='\\' || c=='/') break;
		++it;
	}
	ret.insert(ret.begin(), path.rbegin(), it);
	std::reverse(ret.begin(), ret.end());
	return ret;
}

bool WinrtFileMgr::isSpaceFree(unsigned long nBytes)
{
	PULARGE_INTEGER pFreeSpace = new ULARGE_INTEGER;
	
	auto path = ApplicationData::Current->LocalFolder->Path->Data();
	if(GetDiskFreeSpaceExW(path, pFreeSpace, NULL, NULL)) {
		if(pFreeSpace->HighPart > 0) return true;
		if(pFreeSpace->LowPart >= nBytes) return true;
		else return false;
	}

	// this code shouldn't execute but if the space cannot be retrieved, pass..
	return true;
}

/*
// Split a given file in multiple files of size "chunkSize".
// "file.ext" will be split in "file.ext.part0", "file.ext.part1" etc...
void WinrtFileMgr::splitFile(std::string path, u32 chunkSize)
{
	char outputFilename[512];
	char* buffer = (char*)malloc(chunkSize);
	
	FILE* fileToRead;
	fopen_s(&fileToRead, path.c_str(), "rb");
	if(fileToRead != NULL) 
	{
		PastaLogDebug("DLC load file: %s\n", path.c_str());

		// get file size
		fseek(fileToRead , 0 , SEEK_END);
		long remainingBytes = ftell(fileToRead);
		rewind(fileToRead);

		// split the file
		int i = 0;
		size_t realChunkSize = 0;
		size_t nextChunkSize = chunkSize;
		if(remainingBytes < chunkSize) nextChunkSize = remainingBytes;

		while(1)
		{
			sprintf(outputFilename, "%s.part%d", path.c_str(), i);

			fread(buffer, nextChunkSize, 1, fileToRead);

			FILE* outputFile = fopen(outputFilename, "wb");
			fwrite(buffer, nextChunkSize, 1, outputFile);
			fclose(outputFile);
			++i;

			remainingBytes -= nextChunkSize;

			if(remainingBytes <= 0) break;
			if(remainingBytes < chunkSize) nextChunkSize = remainingBytes;
			PastaLogDebug("Remaining bytes: %d, Next chunk size: %d\n", remainingBytes, nextChunkSize);
		}
	}

	fclose(fileToRead);
	free(buffer);
}

// Reconstruct a file from its split parts.
// Giving "file.ext" to this function will look for any file named "file.ext.part0", "file.ext.part1" etc... in the same location.
// These ".part" files will be merged into a single file named "file.ext".
void WinrtFileMgr::unsplitFile(std::string path)
{
	char partFilename[512];
	char buffer[256];
	FILE* outputFile = fopen(path.c_str(), "wb");
	int i = 0;

	while(1)
	{
		sprintf(partFilename, "%s.part%d", path.c_str(), i);

		FILE* partFile = fopen(partFilename, "rb");
		if(partFile)
		{
			// get file size
			fseek(partFile, 0, SEEK_END);
			long nbBytes = ftell(partFile);
			rewind(partFile);

			// copy file content
			char* buffer = (char*)malloc(nbBytes);
			fread(buffer, nbBytes, 1, partFile);
			fwrite(buffer, nbBytes, 1, outputFile);

			fclose(partFile);
			free(buffer);

			PastaLogDebug("Unsplit file %s\n", partFilename);
		}
		else break;

		++i;
	}

	fclose(outputFile);
}
*/